from abc import ABC, abstractmethod
import asyncio
import uuid
from typing import Any, Callable, Literal, TypedDict


QueueStatus = Literal["waiting", "running", "completed", "failed"]
JobPriority = Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
JOB_PRIORITIES: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


class IJob(TypedDict):
    identifier: str
    priority: JobPriority
    name: str
    status: QueueStatus
    args: tuple[Any]
    kwargs: dict[str, Any]


class QueueKey:
    def __init__(
        self,
        name: str,
        status: QueueStatus = "waiting",
        priority: int = 9,
        identifier: str | None = None,
    ) -> None:
        self.priority = priority
        self.status = status
        self.name = name
        self.identifier = identifier if identifier else str(uuid.uuid4())

    async def update(self, status: QueueStatus, priority: int = 9):
        self.status = status
        self.priority = priority

    @property
    def code(self):
        return f"{self.priority}:{self.status}:{self.name}:{self.identifier}"


class GetJobsResult(TypedDict):
    identifier: str
    key: str
    priority: JobPriority
    status: QueueStatus
    kwargs: dict[str, Any]
    args: tuple[Any, ...]


class IFQueueStorageAdapter(ABC):
    @abstractmethod
    async def add_job(self, key: QueueKey, priority: JobPriority = 9, *args, **kwargs):
        ...

    @abstractmethod
    async def get_jobs(
        self, name: str, limit: int, status: QueueStatus
    ) -> list[GetJobsResult]:
        ...

    @abstractmethod
    async def complete_job(self, name: str, key: QueueKey):
        ...

    @abstractmethod
    async def update_status(
        self,
        key: QueueKey,
        name: str,
        status: QueueStatus,
        priority: JobPriority | None = None,
    ):
        """ジョブのステータスを更新します

        Parameters
        ----------
        key : QueueKey
            キー
        name : str
            ジョブキューの名前
        status : QueueStatus
            新しいステータス
        """
        ...

    @abstractmethod
    async def count_jobs(self, name: str, status: QueueStatus) -> int:
        """指定された状態のジョブの総数を返します

        Parameters
        ----------
        name : str
            キューの名前
        status : QueueStatus
            ジョブの状態

        Returns
        -------
        int
            ジョブの総数
        """
        ...


class Queue:
    def __init__(
        self,
        name: str,
        queue_storage_adapter: IFQueueStorageAdapter,
        func: Callable | None = None,
        *,
        success_func: Callable | None = None,
        max_process: int = 5,
        max_retry: int = 5,
        cooldown: int = 5,
    ) -> None:
        self.is_startup: bool = True
        self.name = name
        self.func = func
        self.success_func = success_func
        self.max_process: int = max_process
        self.current_run_process_number: int = 0
        self.max_retry = max_retry
        self.cooldown = cooldown
        self.next_stop = False
        self.queue_storage_adapter: IFQueueStorageAdapter = queue_storage_adapter
        self.loop = asyncio.get_event_loop()

    def run(self) -> asyncio.Task[None]:
        if self.func is None:
            raise Exception("funcが定義されていません")
        task = self.loop.create_task(self.scheduler())
        return task

    def stop(self) -> None:
        self.next_stop = True

    async def count_jobs(self, status: QueueStatus = "waiting"):
        return await self.queue_storage_adapter.count_jobs(
            name=self.name, status=status
        )

    async def add(self, priority: JobPriority = 9, *args, **kwargs):
        # キーは 状態:キューの名前:uuid で構成される
        await self.queue_storage_adapter.add_job(
            QueueKey(self.name, status="waiting", priority=priority),
            *args,
            **kwargs,
        )

    async def get_queues(self, status: QueueStatus):
        return await self.queue_storage_adapter.get_jobs(
            name=self.name,
            limit=self.max_process - self.current_run_process_number,
            status=status,
        )

    async def count_running_process(self):
        return await self.count_jobs(status="running")

    async def scheduler(self):
        while True:
            if self.next_stop:
                return

            running_process_number = await self.count_running_process()
            if running_process_number <= self.max_process or self.is_startup:
                allow_add_job_number = self.max_process - running_process_number
                if self.is_startup:
                    jobs = await self.get_queues("running")
                    if len(jobs) == 0:
                        jobs = await self.get_queues("waiting")
                        self.is_startup = False
                else:
                    jobs = await self.get_queues("waiting")
                count = 1
                for job in jobs:
                    if (
                        count <= allow_add_job_number or self.is_startup
                    ):  # 現在実行中のプロセスが最大プロセス以下か確認
                        if self.is_startup and running_process_number == 0:
                            self.is_startup = False
                        self.loop.create_task(
                            self.task(
                                QueueKey(
                                    name=self.name,
                                    status=job["status"],
                                    priority=job["priority"],
                                    identifier=job["identifier"],
                                ),
                                *job["args"],
                                **job["kwargs"],
                            ),
                            name=f'queue: {job["key"]}',
                        )
                        count = count + 1
                    # continue

            await asyncio.sleep(self.cooldown)

    async def task(self, key: QueueKey, *args, **kwargs):
        if self.func is None:
            raise Exception("funcが定義されていません")
        await self.queue_storage_adapter.update_status(key, self.name, "running")
        try:
            await self.func(*args, **kwargs)
        except Exception as e:
            await self.queue_storage_adapter.update_status(key, self.name, "failed")
            return

        await self.queue_storage_adapter.update_status(key, self.name, "completed")
        if self.success_func:
            await self.success_func(self.name, key, *args, **kwargs)
        # 処理終了後
        await self.queue_storage_adapter.complete_job(self.name, key)
        self.current_run_process_number = self.current_run_process_number - 1
        return



import json
from typing import Any
from catline.queue import (
    JOB_PRIORITIES,
    GetJobsResult,
    IFQueueStorageAdapter,
    IJob,
    JobPriority,
    QueueKey,
    QueueStatus,
)

try:
    import redis.asyncio  # type: ignore
except ModuleNotFoundError:
    HAS_REDIS = False
else:
    HAS_REDIS = True


class QueueStorageRedisAdapter(IFQueueStorageAdapter):
    def __init__(self, **kwargs) -> None:
        if HAS_REDIS is False:
            raise Exception("Redis has not installed.")
        self.redis_connection = redis.asyncio.Redis(**kwargs)

    async def add_job(self, key: QueueKey, *args, **kwargs):
        data: IJob = {
            "args": args,
            "identifier": key.identifier,
            "kwargs": kwargs,
            "name": key.name,
            "priority": key.priority,
            "status": key.status,
        }  # type: ignore
        await self.redis_connection.set(key.code, json.dumps(data, ensure_ascii=False))

    async def get_jobs(
        self, name: str, limit: int, status: QueueStatus
    ) -> list[GetJobsResult]:
        result: list[GetJobsResult] = []
        for priority in JOB_PRIORITIES:
            if len(result) == limit:
                break
            async for _key in self.redis_connection.scan_iter(
                f"{priority}:{status}:{name}:*"
            ):
                key = _key.decode("utf-8")
                value: str | None = await self.redis_connection.get(key)

                assert value  # 普通あるはず

                _value: IJob = json.loads(value)
                result.append(
                    GetJobsResult(
                        key=key,
                        args=_value["args"],
                        kwargs=_value["kwargs"],
                        status=_value["status"],
                        identifier=_value["identifier"],
                        priority=_value["priority"],
                    )
                )
        return result

    async def complete_job(self, name: str, key: QueueKey):
        ...
        # await redis_connection.delete(key)

    async def count_jobs(self, name: str, status: QueueStatus) -> int:
        count = 0
        keys = []
        async for key in self.redis_connection.scan_iter(f"*:{status}:{name}:*"):
            count = count + 1
            keys.append(key)
        return count

    async def update_status(
        self,
        key: QueueKey,
        name: str,
        status: QueueStatus,
        priority: JobPriority | None = None,
    ):
        new_priority = priority if priority else key.priority
        raw_data = await self.redis_connection.get(key.code)
        await self.redis_connection.delete(key.code)
        if raw_data is None:
            raise Exception(f"指定されたジョブ: {key.code} は存在しません")
        data: IJob = json.loads(raw_data)
        data["priority"] = new_priority  # type: ignore
        data["status"] = status
        await key.update(status=status, priority=new_priority)  # type: ignore
        await self.redis_connection.set(key.code, json.dumps(data, ensure_ascii=False))



import os
from typing import TypedDict

from catline.utils import catline_json as json
from catline.queue import (
    JOB_PRIORITIES,
    GetJobsResult,
    IFQueueStorageAdapter,
    IJob,
    JobPriority,
    QueueKey,
    QueueStatus,
)


class IJobs(TypedDict):
    priority_0: dict[str, IJob]
    priority_1: dict[str, IJob]
    priority_2: dict[str, IJob]
    priority_3: dict[str, IJob]
    priority_4: dict[str, IJob]
    priority_5: dict[str, IJob]
    priority_6: dict[str, IJob]
    priority_7: dict[str, IJob]
    priority_8: dict[str, IJob]
    priority_9: dict[str, IJob]


class IQueues(TypedDict):
    jobs: IJobs


class QueueStorageJSONAdapter(IFQueueStorageAdapter):
    def __init__(self) -> None:
        if os.path.exists("./queue.json"):
            with open("./queue.json", mode="r", encoding="utf-8") as f:
                queues: IQueues = json.load(f)
        else:
            queues: IQueues = {
                "jobs": {
                    "priority_0": {},
                    "priority_1": {},
                    "priority_2": {},
                    "priority_3": {},
                    "priority_4": {},
                    "priority_5": {},
                    "priority_6": {},
                    "priority_7": {},
                    "priority_8": {},
                    "priority_9": {},
                }
            }
        self.queues = queues

    async def save_file(self):
        with open("./queue.json", mode="w", encoding="utf-8") as f:
            json.dump(self.queues, f, ensure_ascii=False, indent=4)

    async def add_job(self, key: QueueKey, *args, **kwargs):
        self.queues["jobs"][f"priority_{key.priority}"][key.identifier] = {
            "identifier": key.code,
            "priority": key.priority,
            "name": key.name,
            "status": key.status,
            "args": args,
            "kwargs": kwargs,
        }
        await self.save_file()

    async def get_jobs(
        self, name: str, limit: int, status: QueueStatus
    ) -> list[GetJobsResult]:
        result: list[GetJobsResult] = []
        for priority in JOB_PRIORITIES:
            if len(result) == limit:
                break
            for key, job in self.queues["jobs"][f"priority_{priority}"].items():  # TODO: JSONの形式を変更する。 nameの中に優先度を作成することで、他のジョブをforから除外できる
                _status = job["status"]
                _name = job["name"]
                if priority == priority and status == _status and name == _name:
                    _args = job["args"]
                    _kwargs = job["kwargs"]
                    _status = job["status"]
                    result.append(GetJobsResult(key=key, args=_args, kwargs=_kwargs, identifier=key, status=_status, priority=priority))  # type: ignore
                    if len(result) == limit:
                        break
                continue
        return result

    async def remove_by_key(self, key_id: str, priority: int):
        del self.queues["jobs"][f"priority_{priority}"][key_id]
        await self.save_file()

    async def complete_job(self, name: str, key: QueueKey):
        completed_number = await self.count_jobs(name, "completed")
        if completed_number > 50:
            completed_jobs = await self.get_jobs(name, 20, "completed")
            for job in completed_jobs:
                await self.remove_by_key(job["identifier"], priority=job["priority"])

    async def count_jobs(self, name: str, status: QueueStatus) -> int:
        count = 0
        for priority in JOB_PRIORITIES:
            for job in self.queues["jobs"][f"priority_{priority}"].values():
                if job["status"] == status and job["name"] == name:
                    count = count + 1
        return count

    async def update_status(
        self,
        key: QueueKey,
        name: str,
        status: QueueStatus,
        priority: JobPriority | None = None,
    ):
        new_priority = priority or key.priority
        new_key = await key.update(status=status, priority=new_priority)
        self.queues["jobs"][f"priority_{key.priority}"][key.identifier][
            "status"
        ] = status
        await self.save_file()



import json


try:
    import orjson  # type: ignore
except ModuleNotFoundError:
    HAS_ORJSON = False
else:
    HAS_ORJSON = True
    
if HAS_ORJSON:
    catline_json = orjson  # type: ignore
else:
    catline_json = json



import pathlib

from setuptools import setup

import versioneer

description = 'Simple job queue'
readme_file = pathlib.Path(__file__).parent / 'README.md'
with readme_file.open(encoding='utf-8') as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as f:
    requirements = f.read().splitlines()

extras_require = {
    'redis': ['redis'],
}

packages = [
    'catline',
]

setup(
    name='catline',
    version='0.0.1',
    install_requires=requirements,
    url='https://github.com/yupix/catline',
    author='yupix',
    author_email='yupi0982@outlook.jp',
    license='MIT',
    python_requires='>=3.11, <4.0',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=packages,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python :: 3.11',
        'Natural Language :: Japanese',
        'License :: OSI Approved :: MIT License',
    ],
    extras_require=extras_require,
)